#include "particule.h"


// Compute E*qP/norm(qP)
// Return false if norm(qP) < eps
bool compute_e(charge_t c, vec p, double eps, vec *e)
{
  vec qp = vec_sub(p, c.pos); // Distance entre les deux particules
  double norm_qp = vec_norm(qp);
  // printf("norm = %f\r\n", norm_qp);
  if(norm_qp < eps)
  {
    return false;
  }

  double cst_e = (K * c.q) / pow(norm_qp,2);// E

  e->x = (cst_e * qp.x) / norm_qp;
  e->y = (cst_e * qp.y) / norm_qp;
  return true;
}


// Compute the normalized sum of Ei*qiP/norm(qiP)
// Return false if for some qiP, norm(qiP) < eps
bool compute_total_normalized_e(charge_t *charges, int num_charges, vec p, double eps, vec *e)
{
  vec tmp_e;
  for (int i = 0; i < num_charges; i++)
  {
    if(compute_e(charges[i], p, eps, &tmp_e) == false)//si pas faux, continue les calcul
    {
      return false;
    }
    e->x += tmp_e.x;
    e->y += tmp_e.y;
  }
  return true;
}


// Compute and then draw all the points belonging to a field line,
// starting from pos0.
// Returns false if pos0 is not a valid position
// (for example if pos0 is too close to a charge).
bool draw_field_line(struct gfx_context_t *ctxt, charge_t *charges,int num_charges, double dx, vec pos0, double x0,double x1, double y0, double y1, uint32_t max_proximity)
{
  if(draw_field_line_dir(ctxt, charges, num_charges, dx, pos0, x0, x1, y0, y1, true, max_proximity) == false)
  {
    return false;
  }

  if(draw_field_line_dir(ctxt, charges, num_charges, dx, pos0, x0, x1, y0, y1, false, max_proximity) == false)
  {
    return false;
  }
  return true;
}

bool draw_field_line_dir(struct gfx_context_t *ctxt, charge_t *charges,int num_charges, double dx, vec pos0, double x0,double x1, double y0, double y1, bool dir, uint32_t max_proximity)
{
  vec p;
  vec pSuivant;
  vec e_tmp = {.x = 0.0, .y = 0.0};
  bool lower_eps;
  double norm_tmp;
  if(isTooCloseToACharge(charges, num_charges, pos0, max_proximity, ctxt, x0, x1, y0, y1))
  {
    return false;
  }
  //Etape 2: on assigne P = P0
  p = pos0;
  do
  {
    //Etape 3:
    //on calcul Psuivant
    lower_eps = compute_total_normalized_e(charges, num_charges, p, MIN_NORM_E, &e_tmp);//calcul de E
    if(lower_eps == false)
    {
      break;
    }
    // printf("e_tmp {.pos.x = %f, .pos.y = %f};\r\n",e_tmp.x, e_tmp.y);

    norm_tmp = vec_norm(e_tmp);//calcul de la norm de E
    // printf("norm_tmp = %f\r\n",norm_tmp);

    e_tmp = vec_mul(e_tmp, dx);//multiplie E par dx = 1.0/sqrt((context->width*context->width)+(context->height*context->height));
    // printf("e_tmp {.pos.x = %f, .pos.y = %f};\r\n",e_tmp.x, e_tmp.y);

    pSuivant = vec_div(e_tmp, norm_tmp);//on divise E / ||E||   --> pSuivant.x, pSuivant.y
    // printf("pSuivant {.pos.x = %f, .pos.y = %f};\r\n",pSuivant.x, pSuivant.y);

    if(dir) //Direction en avant ou Direction en arrière
    {
      pSuivant = vec_add(p, pSuivant);//ajoute P pour obtenir Psuivant
    }
    else
    {
      pSuivant = vec_sub(p, pSuivant);//ajoute P pour obtenir Psuivant
    }

    // printf("pSuivant {.pos.x = %f, .pos.y = %f};\r\n\r\n",pSuivant.x, pSuivant.y);
    //Etape 4: si P et Psuivant sont a une distance...
    if(isValidCoordinate(p))
    {
      // printf("COUCOU\r\n");
      if(isValidCoordinate(pSuivant))
      {

        if(isCloser(p, pSuivant, OFFSET_TRACE_LINE)==false)//si pas trop proche
        {
          if(isTooCloseToACharge(charges, num_charges, p, max_proximity, ctxt, x0, x1, y0, y1))
          {
            break;
          }
          coordinates_t p0  = position_to_coordinates(ctxt->width, ctxt->height, x0, x1, y0, y1, p); 
          coordinates_t p1  = position_to_coordinates(ctxt->width, ctxt->height, x0, x1, y0, y1, pSuivant);
          gfx_draw_line(ctxt, p0, p1, COLOR_YELLOW);
        }
      }
      else
      {
        break;//quitte la boucle
      }
    }
    else
    {
      break;//quitte la boucle
    }
    //Etape 5: on assigne P = Psuivant
    p = pSuivant;
    // printf("ICI\r\n");
  } while(lower_eps);
  return true;
}

bool isTooCloseToACharge(charge_t *charges, int num_charges, vec pos0, double distance, struct gfx_context_t *ctxt, double x0,double x1, double y0, double y1)
{
  for (int i = 0; i < num_charges; i++)
  {
    if(isCloserCoordinate(charges[i].pos, pos0, distance, ctxt, x0, x1, y0, y1))
    {
      return true;
    }
  }
  return false;
}

bool isCloserCoordinate(vec pos1, vec pos2, double distance, struct gfx_context_t *ctxt, double x0,double x1, double y0, double y1)
{
  coordinates_t p0, p1;
  p0 = position_to_coordinates(ctxt->width, ctxt->height, x0, x1, y0, y1, pos1);
  p1 = position_to_coordinates(ctxt->width, ctxt->height, x0, x1, y0, y1, pos2);
  if(abs((int32_t)p0.row - (int32_t)p1.row) < distance)
  {
    if(abs((int32_t)p0.column - (int32_t)p1.column) < distance)
    {
      return true;
    }
  }
  return false;
}

//Controle si les deux points dans l'espace sont plus proche de la distance(Parametre)
bool isCloser(vec pos1, vec pos2, double distance)
{
  if(fabs(pos1.x - pos2.x) < distance)
  {
    return true;
  }
  if(fabs(pos1.y - pos2.y) < distance)
  {
    return true;
  }
  return false;
}

bool isValidCoordinate(vec pos)
{
  if(pos.x < 0)
  {
    return false;
  }
  if(pos.x > 1)
  {
    return false;
  }
  if(pos.y < 0)
  {
    return false;
  }
  if(pos.y > 1)
  {
    return false;
  }
  return true;
}

// Draw all the charges
// A circle with minus sign for negative charges
// A circle with a plus sign for positive charges
void draw_charges(struct gfx_context_t *context, charge_t *charges,int num_charges, double x0, double x1, double y0, double y1, uint32_t num_particule, uint32_t max_proximity)
{
  coordinates_t c;
  coordinates_t p0, p1;
  int radius = RADIUS_CHARGE;
  for(int i=0;i<num_charges;i++)
  {
    c = position_to_coordinates(context->width, context->height, x0, x1, y0, y1, charges[i].pos);
    if(charges[i].q > 0)//la charge est positive
    {
      p0.row = c.row - ((radius/2)+1);
      p0.column = c.column;
      p1.row = c.row + ((radius/2)+1);
      p1.column = c.column;
      gfx_draw_circle(context, c, radius ,COLOR_RED);
      gfx_draw_line(context, p0, p1, COLOR_RED);
      p0.row = c.row;
      p0.column = c.column - ((radius/2)+1);
      p1.row = c.row;
      p1.column = c.column + ((radius/2)+1);
      gfx_draw_line(context, p0, p1, COLOR_RED);
    }
    else
    {
      p0.row = c.row - ((radius/2)+1);
      p0.column = c.column;
      p1.row = c.row + ((radius/2)+1);
      p1.column = c.column;
      gfx_draw_circle(context, c, radius ,COLOR_BLUE);
      gfx_draw_line(context, p0, p1, COLOR_BLUE);
    }
  }
  double dx = 1.0/sqrt((context->width*context->width)+(context->height*context->height));
  vec pos0;
  uint32_t essai;//pour eviter de tourner en boucle
  for (uint32_t i = 0; i < num_particule; i++)
  {
    essai = 0;
    do
    {
      pos0.x = rand_one();
      pos0.y = rand_one();
      essai++;
      // printf("ESSAI = %d\r\n", essai);
      if(essai >= 100)
      {
        // printf("HELLO\r\n");
        essai = 0;
        break;
      }
      // printf("pos0 {.pos.x = %f, .pos.y = %f};\r\n", pos0.x, pos0.y);
    } while(draw_field_line(context, charges, num_charges, dx, pos0, x0, x1, y0, y1, max_proximity) == 0);
  }

}

// https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm#:~:text=Bresenham's%20line%20algorithm%20is%20a,straight%20line%20between%20two%20points.
void gfx_draw_line(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color)
{
  if(abs((int)p1.column - (int)p0.column) < abs((int)p1.row - (int)p0.row))
  {
    if(p0.row > p1.row)
    {
      drawLineLow(ctxt, p1, p0, color);
    }
    else
    {
      drawLineLow(ctxt, p0, p1, color);
    }
  }
  else
  {
    if(p0.column > p1.column)
    {
      drawLineHigh(ctxt, p1, p0, color);
    }
    else
    {
      drawLineHigh(ctxt, p0, p1, color);
    }
  }
}

void drawLineLow(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color)
{
  int dx = p1.row - p0.row;//x1 - x0
  int dy = p1.column - p0.column;//y1 - y0
  int d = (2*dy) - dx;
  int y = p0.column;
  int yi = 1;
  if(dy < 0)
  {
    yi = -1;
    dy=-dy;
  }
  for (unsigned int x = p0.row; x < p1.row; x++)
  {
    gfx_putpixel(ctxt, x, y, color);
    if(d > 0)
    {
      y+=yi;
      d = d + (2*(dy-dx));
    }
    else
    {
      d = d + (2*dy);
    }
  }
}

void drawLineHigh(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color)
{
  int dx = p1.row - p0.row;//x1 - x0
  int dy = p1.column - p0.column;//y1 - y0
  int d = (2*dx) - dy;
  int x = p0.row;
  int xi = 1;
  if(dx < 0)
  {
    xi = -1;
    dx=-dx;
  }
  for (unsigned int y = p0.column; y < p1.column ; y++)
  {
    gfx_putpixel(ctxt, x, y, color);
    if(d > 0)
    {
      x+=xi;
      d = d + (2*(dx-dy));
    }
    else
    {
      d = d + (2*dx);
    }
  }
}

//Algorithme de tracage de cercle
void gfx_draw_circle(struct gfx_context_t *ctxt, coordinates_t c, uint32_t r, uint32_t color)
{

  int f = 1 - r;
  int ddF_x = 1;
  int ddF_y = -2 * r;
  int x = 0;
  int y = r;

  int temp_x0_p_x;
  int temp_x0_n_x;
  int temp_x0_p_y;
  int temp_x0_n_y;

  int temp_y0_p_x;
  int temp_y0_n_x;
  int temp_y0_p_y;
  int temp_y0_n_y;

  gfx_putpixel(ctxt, c.row, c.column + r, color);
  gfx_putpixel(ctxt, c.row, c.column - r, color);
  gfx_putpixel(ctxt, c.row + r, c.column, color);
  gfx_putpixel(ctxt, c.row - r, c.column, color);

  while (x < y)
  {
      if (f >= 0)
      {
        y--;
        ddF_y += 2;
        f += ddF_y;
      }
      x++;
      ddF_x += 2;
      f += ddF_x;

      //pour eviter de refaire les calcul plusieur fois
      temp_x0_p_x = c.row + x;
      temp_x0_n_x = c.row - x;
      temp_x0_p_y = c.row + y;
      temp_x0_n_y = c.row - y;

      temp_y0_p_x = c.column + x;
      temp_y0_n_x = c.column - x;
      temp_y0_p_y = c.column + y;
      temp_y0_n_y = c.column - y;

      gfx_putpixel(ctxt, temp_x0_p_x, temp_y0_p_y, color);
      gfx_putpixel(ctxt, temp_x0_n_x, temp_y0_p_y, color);
      gfx_putpixel(ctxt, temp_x0_p_x, temp_y0_n_y, color);
      gfx_putpixel(ctxt, temp_x0_n_x, temp_y0_n_y, color);
      gfx_putpixel(ctxt, temp_x0_p_y, temp_y0_p_x, color);
      gfx_putpixel(ctxt, temp_x0_n_y, temp_y0_p_x, color);
      gfx_putpixel(ctxt, temp_x0_p_y, temp_y0_n_x, color);
      gfx_putpixel(ctxt, temp_x0_n_y, temp_y0_n_x, color);
  }
}
