#ifndef PARTICULE_H
#define PARTICULE_H

#include <stdio.h>
#include <math.h>
#include <stdlib.h> // pour acceder au random
#include <time.h>   // pour acceder au time machine
#include <stdbool.h> // librairie C pour les boolean
#include "vector.h"
#include "gfx.h"
#include "utils.h"

#define OFFSET_TRACE_LINE 0.000001 //Distance pour tracer le trait
#define MIN_NORM_E 0.0000001  // Epsilone
#define RADIUS_CHARGE 10 // 10 pixel de taille pour la charge

// Compute E*qP/norm(qP)
// Return false if norm(qP) < eps
bool compute_e(charge_t c, vec p, double eps, vec *e);

// Compute the normalized sum of Ei*qiP/norm(qiP)
// Return false if for some qiP, norm(qiP) < eps
bool compute_total_normalized_e(charge_t *charges, int num_charges, vec p, double eps, vec *e);


// Compute and then draw all the points belonging to a field line,
// starting from pos0.
// Returns false if pos0 is not a valid position
// (for example if pos0 is too close to a charge).
bool draw_field_line(struct gfx_context_t *ctxt, charge_t *charges,int num_charges, double dx, vec pos0, double x0,double x1, double y0, double y1, uint32_t max_proximity);

bool draw_field_line_dir(struct gfx_context_t *ctxt, charge_t *charges,int num_charges, double dx, vec pos0, double x0,double x1, double y0, double y1, bool dir, uint32_t max_proximity);
// Draw all the charges
// A circle with minus sign for negative charges
// A circle with a plus sign for positive charges
void draw_charges(struct gfx_context_t *context, charge_t *charges,int num_charges, double x0, double x1, double y0, double y1, uint32_t num_particule, uint32_t max_proximity);
bool isTooCloseToACharge(charge_t *charges, int num_charges, vec pos0, double distance, struct gfx_context_t *ctxt, double x0,double x1, double y0, double y1);
bool isCloser(vec pos1, vec pos2, double distance);
bool isCloserCoordinate(vec pos1, vec pos2, double distance, struct gfx_context_t *ctxt, double x0,double x1, double y0, double y1);
bool isValidCoordinate(vec pos);

void gfx_draw_line(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color);
void drawLineLow(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color);
void drawLineHigh(struct gfx_context_t *ctxt, coordinates_t p0, coordinates_t p1, uint32_t color);
void gfx_draw_circle(struct gfx_context_t *ctxt, coordinates_t c, uint32_t r, uint32_t color);

#endif
